/* This file is part of gacopyz.
   Copyright (C) 2006, 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include <gacopyz_priv.h>

static int cleanup_needed;

static void
print_status(gacopyz_conn_t conn, pid_t pid, int status, int expect_term)
{
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) == 0)
			gacopyz_log(conn, SMI_LOG_DEBUG,
				    "child %lu exited successfully",
				    (unsigned long) pid);
		else
			gacopyz_log(conn, SMI_LOG_ERR,
				    "child %lu failed with status %d",
				    (unsigned long) pid,
				    WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		int prio;
		if (expect_term && WTERMSIG(status) == SIGTERM)
			prio = SMI_LOG_DEBUG;
		else
			prio = SMI_LOG_ERR;
		gacopyz_log(conn, prio,
			    "child %lu terminated on signal %d",
			    (unsigned long) pid, WTERMSIG(status));
	} else if (WIFSTOPPED(status))
		gacopyz_log(conn, SMI_LOG_ERR,
			    "child %lu stopped on signal %d",
			    (unsigned long) pid, WSTOPSIG(status));
#ifdef WCOREDUMP
	else if (WCOREDUMP(status))
		gacopyz_log(conn, SMI_LOG_ERR,
			    "child %lu dumped core",
			    (unsigned long) pid);
#endif
	else
		gacopyz_log(conn, SMI_LOG_ERR,
			    "child %lu terminated with unrecognized status",
			    (unsigned long) pid);
}

static void
cleanup_children(gacopyz_conn_t conn, int expect_term)
{
	if (!cleanup_needed)
		return;
	for (;;) {
		int status;
		pid_t pid = waitpid((pid_t)-1, &status, WNOHANG);
		if (pid <= 0)
			break;
		print_status(conn, pid, status, expect_term);
	}
	cleanup_needed = 0;
}

void
gacopyz_cleanup_children(gacopyz_conn_t conn)
{
	cleanup_children(conn, 0);
}

void
gacopyz_cleanup_conn(gacopyz_conn_t conn)
{
	gacopyz_log(conn, SMI_LOG_DEBUG, "terminating subprocesses");
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	cleanup_children(conn, 1);
}

static RETSIGTYPE
sig_child(int sig)
{
	cleanup_needed = 1;
	signal(sig, sig_child);
}

void
gacopyz_setup_signals()
{
	signal(SIGCHLD, sig_child);
	signal(SIGPIPE, SIG_IGN); /* FIXME */
}
